import sys
import os
from json import JSONDecodeError
from .featurize_client import FeaturizeClient


def _find_token():
    token_from_env = os.getenv('FEATURIZE_API_TOKEN')
    cfg_file = os.getenv('FEATURIZE_CFG_FILE', '/etc/featurize_token.txt')
    if token_from_env:
        return token_from_env
    try:
        with open(cfg_file, 'r') as f:
            return f.read()
    except FileNotFoundError:
        return None
    except JSONDecodeError:
        sys.exit(f'config file {cfg_file} parse error')


def _find_instance_id():
    return os.getenv('FEATURIZE_INSTANCE_ID')


def create_client_from_env():
    token = _find_token()
    instance_id = _find_instance_id()
    if token is None:
        sys.exit('Token is missed')
    return FeaturizeClient(token=token, instance_id=instance_id)


__version__ = '0.0.26'
__all__ = ['FeaturizeClient', 'create_client_from_env']



from .resource import Instance, Dataset, OssCredentials, Notebook, Port, Event, TempToken


class FeaturizeClient:

    def __init__(self, token, instance_id=None):
        self.token = token
        self.instance_id = instance_id

    @property
    def instance(self) -> Instance:
        return self._get_resource(Instance)

    @property
    def event(self) -> Event:
        return self._get_resource(Event)

    @property
    def port(self) -> Port:
        return self._get_resource(Port)

    @property
    def dataset(self) -> Dataset:
        return self._get_resource(Dataset)

    @property
    def oss_credential(self) -> OssCredentials:
        return self._get_resource(OssCredentials)

    @property
    def notebook(self) -> Notebook:
        return self._get_resource(Notebook)

    @property
    def temptoken(self) -> TempToken:
        return self._get_resource(TempToken)

    def _get_resource(self, resource_type):
        resource_name = resource_type.__name__
        if not hasattr(self, resource_name):
            setattr(self, resource_name, resource_type(self.token, self.instance_id))
        return getattr(self, resource_name)



import hashlib
from json.decoder import JSONDecodeError
import requests
import logging
from pathlib import Path
import json
import os
from tqdm import tqdm

logger = logging.getLogger(__file__)


def to_camel_case(snake_str: str) -> str:
    components = snake_str.split("_")
    return components[0] + "".join(x.title() for x in components[1:])


def extract(source, dest):
    out = os.system(f'unzip -o "{source}" -d {dest} >> ~/.server.log 2>&1 ')
    if out == 0:
        return 0

    out = os.system(f'tar -zxf "{source}" -C {dest} >> ~/.server.log 2>&1')
    if out == 0:
        return 0

    out = os.system(f'tar -xf "{source}" -C {dest} >> ~/.server.log 2>&1')
    return out


class HTTPCodeError(Exception):
    def __init__(self, code: str, response: requests.Response):
        super().__init__(f"HTTP request failed with code: {code}, body: {response}")


class ServiceError(Exception):
    def __init__(self, code: str, message: str):
        self.code = code
        self.message = message
        super().__init__(f"Error {code}: {message}")


class Resource:

    base = "https://featurize.cn/bus/api/v1"

    def __init__(self, token: str, instance_id: str = None):
        self.token = token
        self.instance_id = instance_id

    def _http(
        self,
        url: str,
        method: str = "get",
        data: dict = None,
        headers: dict = {},
        **kwargs,
    ) -> requests.Response:
        url = f"{self.base}{url}"
        if method in ["get", "delete", "head"]:
            kwargs = {"params": data, **kwargs}
        else:
            kwargs = {"json": data, **kwargs}
        req = requests.request(
            method, url, headers={"Token": self.token, **headers}, timeout=30, **kwargs
        )

        if req.status_code != 200:
            raise HTTPCodeError(req.status_code, req.json())

        res = req.json()
        if res["status"] != 0:
            raise ServiceError(res["status"], res["message"])

        return res["data"]

    @property
    def cache_dir(self) -> Path:
        from_env = os.getenv("FEATURIZE_CACHE_DIR")
        path = Path(from_env) if from_env else (Path.home() / ".featurize")
        if not path.exists():
            path.mkdir(parents=True, exist_ok=True)
        return path

    def cache_file(self, name: str) -> Path:
        file_path = self.cache_dir / name
        return file_path

    def _check_instance_id(self):
        if self.instance_id is None:
            raise RuntimeError('Instance ID was not found, this command should been run in one of featurize instance')


class Instance(Resource):

    def release(self) -> dict:
        return self._http(f"/virtual_machine/{self.instance_id}/release", "post")


class Notebook(Resource):
    def create(self, file: str, name: str):
        # check cache file, if exists create version instead of notebook
        s = hashlib.sha1()
        s.update(file.encode())
        cache_file_name = f"notebook.{s.hexdigest()}.json"
        cache_file = self.cache_file(cache_file_name)
        upload_files = {"file": open(file, "rb")}
        try:
            notebook_id = json.loads(cache_file.read_text())["id"]
            response = self._http(
                f"/notebooks/{notebook_id}/version",
                "POST",
                params={"name": name},
                files=upload_files,
            )
        except (JSONDecodeError, FileNotFoundError, KeyError):
            response = self._http(
                "/notebooks", "POST", params={"name": name}, files=upload_files
            )
        self.cache_file(cache_file_name).write_text(json.dumps(response))
        return response


class Port(Resource):

    def list(self) -> dict:
        self._check_instance_id()
        return self._http(f"/virtual_machine/{self.instance_id}/exported_ports")

    def create(self, local_port: str) -> dict:
        try:
            self._check_instance_id()
            return self._http(f"/virtual_machine/{self.instance_id}/exported_ports", "POST", params={"local_port": local_port})
        except ServiceError as e:
            if e.code == 10029:
                raise RuntimeError("Maximum number of exported ports reached, please unexport before export new ports")
            if e.code == 10030:
                raise RuntimeError("Can not export / unexport system port 22")

    def destroy(self, local_port: str) -> dict:
        self._check_instance_id()
        try:
            return self._http(
                f"/virtual_machine/{self.instance_id}/exported_ports/{local_port}", "DELETE", params={"local_port": local_port}
            )
        except ServiceError as e:
            if e.code == 10030:
                raise RuntimeError("Can not export / unexport system port 22")


class Dataset(Resource):
    def create(self, name: str, range: str = "private", description: str = "") -> dict:
        return self._http(
            f"/datasets/",
            "post",
            {"name": name, "description": description, "range": range},
        )

    def update(self, dataset_id: str, **kwargs) -> dict:
        return self._http(f"/datasets/{dataset_id}", "patch", kwargs)

    def download(self, dataset_id: str):
        dataset = self._http(f"/datasets/{dataset_id}")
        if not dataset["uploaded"]:
            raise RuntimeError("Dataset uploaded")
        if dataset["cache_progress"] != 100:
            raise RuntimeError("Dataset is not synced yet, please try later")

        chunk_size = 1024 * 1024 * 4
        dataset_dir = Path.home() / "data"
        dataset_dir.mkdir(parents=True, exist_ok=True)
        dataset_file = dataset_dir / dataset["path"].split("/")[-1]
        url = (
            f'http://{dataset["cache"][os.getenv("INSTANCE_REGION")]}/{dataset["path"]}'
        )

        with requests.get(url, stream=True, proxies={"http": "", "https": ""}) as resp:
            resp.raise_for_status()
            file_size = int(resp.headers["content-length"])
            progress_bar = tqdm(total=file_size, unit="iB", unit_scale=True)
            with open(dataset_file, "wb") as data_fd:
                for chunk in resp.iter_content(chunk_size=chunk_size):
                    progress_bar.update(len(chunk))
                    data_fd.write(chunk)
            progress_bar.close()
        print("🍬  下载完成，正在解压...")
        result = extract(
            dataset_file.resolve().as_posix(), dataset_dir.resolve().as_posix()
        )
        if result != 0:
            print("🐛  解压失败，请尝试手动解压或查看文件完整性")
        else:
            print("🏁  数据集已经成功添加")


class Event(Resource):

    def create(self, title: str, content: str):
        return self._http(
            f"/virtual_machine/{self.instance_id}/custom_events",
            "post",
            {"title": title, "content": content}
        )


class TempToken(Resource):
    def get(self) -> dict:
        return self._http(f"/temp_token")


class OssCredentials(Resource):
    def get(self) -> dict:
        return self._http(f"/oss_credentials")



import json
from json.decoder import JSONDecodeError
import os
import sys
import uuid
import time

import click
from click.types import Choice
from featurize import create_client_from_env
import wcwidth as _  # noqa
from tabulate import tabulate
from pathlib import Path
import hashlib
import oss2
from tqdm import tqdm
import shutil
from simple_uploader import SimpleUploader

from .featurize_client import FeaturizeClient
from .resource import ServiceError

client : FeaturizeClient = None


@click.group()
@click.option('-t', '--token', required=False,
              help='Your api token')
def cli(token=None):
    global client
    client = create_client_from_env()


@cli.group()
def event():
    pass


@event.command()
@click.option("-t", "--title", required=True)
@click.option("-c", "--content", required=True)
def create(title, content):
    client.event.create(title, content)


@cli.group()
def instance():
    pass


@instance.command()
def release():
    client.instance.release()


@cli.group()
def port():
    pass


@port.command()
@click.option("-f", "--format", default="tabulate")
def list(format):
    ports = client.port.list()
    if format == "tabulate":
        print(tabulate(ports, headers=("本地端口", "外部端口"), tablefmt='grid'))
    elif format == "json":
        print(json.dumps(ports))


@port.command()
@click.argument("local_port")
@click.option("-r", "--raw", is_flag=True, default=False)
def export(local_port, raw):
    new_port = client.port.create(local_port)
    if raw:
        print(new_port, end="")
    else:
        print(f"Local port {local_port} has been exported to {new_port}")
        print(f"You can visit http://workspace.featurize.cn:{new_port} if it's a http server")


@port.command()
@click.argument("local_port")
def unexport(local_port):
    client.port.destroy(local_port)
    print("done")


@cli.group()
def dataset():
    pass


@dataset.command()
@click.argument('file')
@click.option('-n', '--name', default='')
@click.option('-r', '--range', type=Choice(['public', 'personal']), default='personal')
@click.option('-d', '--description', default='')
@click.option('-d', '--proxy', is_flag=True, default=False)
def upload(file, name, range, description, proxy):
    while True:
        try:
            if not proxy:
                os.environ['HTTP_PROXY'] = ''
                os.environ['HTTPS_PROXY'] = ''
                os.environ['ALL_PROXY'] = ''
                os.environ['https_proxy'] = ''
                os.environ['http_proxy'] = ''
                os.environ['all_proxy'] = ''
            if not (file.endswith(".zip") or file.endswith(".tar.gz")):
                sys.exit('Error: uploading file should be one of .zip or .tar.gz type')
            filepath = Path(file)
            total_size = filepath.stat().st_size
            if not filepath.exists():
                sys.exit('Error: file not exists')
            name = name or filepath.name
            sha1 = hashlib.sha1()
            sha1.update((f"{filepath.name}-{total_size}").encode())
            digest = sha1.hexdigest()

            dataset_file = Path.home() / '.featurize' / f"file_upload_checkpoint_{digest}" / "dataset.json"
            dataset_file.parent.mkdir(parents=True, exist_ok=True)
            try:
                dataset = json.loads(dataset_file.read_text())
            except (FileNotFoundError, json.JSONDecodeError):
                # init dataset
                res = client.dataset.create(name, range, description)
                dataset = {
                    'id': res['id'],
                    'dataset_center': res['dataset_center'],
                    'uploader_id': res['uploader_id'],
                    'consumed_bytes': 0
                }
                dataset_file.write_text(json.dumps(dataset))

            token = client.temptoken.get()["token"]
            bar = tqdm(total=100, desc="🚀 正在上传")
            def on_progress(progress):
                bar.update(int(progress["finished_slice"] / progress["all_slice"] * 100) - bar.n)

            su = SimpleUploader(filepath, {
                "endpoint": "https://nas.featurize.cn/users/files",
                "on_progress": on_progress,
                "prefix": f"{os.environ['FEATURIZE_USER_ID']}_{dataset['id']}",
                "headers": {
                    "Authorization": token,
                    "User-Id": os.environ['FEATURIZE_USER_ID'],
                }
            })
            su.upload()
            res = su.checksum()
            bar.update(100)
            bar.close()
            client.dataset.update(
                dataset_id=dataset['id'],
                uploaded=True,
                cache_progress=100,
                path=f"{dataset['uploader_id']}_{dataset['id']}/{filepath.name}",
                size=total_size,
                filename=filepath.name
            )

            shutil.rmtree(dataset_file.parent)

            if res.failed_count == 0:
                print(f"✅ 已成功上传！下载命令：featurize dataset download {dataset['id']}")
            su.clear_meta()
            break

        except Exception as e:
            print(e)
            time.sleep(5)
            pass


@dataset.command()
@click.argument('id')
def download(id):
    try:
        id = uuid.UUID(hex=id)
        client.dataset.download(dataset_id=str(id))
    except ValueError:
        sys.exit("Error: id is not valid")
